# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------

import csv
import os

import pytest
import six


class CSVReporter(object):
    def __init__(self,
                 csv_path,
                 columns,
                 delimiter,
                 quote_char):
        self._csv_path = csv_path
        self._columns = columns
        self._delimiter = delimiter
        self._quote_char = quote_char
        self._rows = []

    @pytest.mark.hookwrapper
    def pytest_runtest_makereport(self, item, call):
        outcome = yield
        report = outcome.get_result()

        if report.when != 'call' and report.passed:
            return

        report._csv_rows = {column_id: dict(column(item, report)) if callable(column) else {column_id: str(column)}
                            for column_id, column in six.iteritems(self._columns)}

    def pytest_runtest_logreport(self, report):
        rows = getattr(report, '_csv_rows', None)
        if rows is not None:
            self._rows.append(rows)
        elif report.when == '???':
            # A worker crashed, and xdist called logreport with a fake report the master generated
            rows = {}
            for column_id, column in six.iteritems(self._columns):
                if callable(column):
                    try:
                        column_val = dict(column(None, report))  # The item isn't available - the master doesn't run collection
                    except Exception:
                        column_val = {}
                else:
                    column_val = {column_id: str(column)}
                rows[column_id] = column_val
            self._rows.append(rows)

    def pytest_sessionfinish(self, session):
        if not self._rows or hasattr(session.config, 'slaveinput'):
            return

        directory = os.path.dirname(self._csv_path)
        if directory and not os.path.exists(directory):
            os.makedirs(directory)

        headers = {column_id: list(sorted(set(header
                                              for row in self._rows
                                              for header in six.iterkeys(row[column_id]))))
                   for column_id in six.iterkeys(self._columns)}

        with open(self._csv_path, 'w', newline='') as out:
            writer = csv.writer(out, delimiter=self._delimiter, quotechar=self._quote_char, quoting=csv.QUOTE_MINIMAL)

            writer.writerow([header
                             for column_id in six.iterkeys(self._columns)
                             for header in headers[column_id]])

            for row in self._rows:
                writer.writerow([row[column_id].get(header, '')
                                 for column_id in six.iterkeys(self._columns)
                                 for header in headers[column_id]])

        session.config.hook.pytest_csv_written(csv_path=self._csv_path)

    def pytest_terminal_summary(self, terminalreporter):
        terminalreporter.write_sep('-', 'CSV report: %s' % self._csv_path)



# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------

from collections import OrderedDict

from . import _hooks
from ._reporter import CSVReporter
from .column import *


def pytest_addoption(parser):
    group = parser.getgroup('terminal reporting')
    group.addoption(
        '--csv',
        dest='csv_path',
        action='store',
        metavar='path',
        default=None,
        help='create CSV report file at given path'
    )
    group.addoption(
        '--csv-columns',
        dest='csv_columns',
        action='store',
        type=str,
        nargs='+',
        default=[ID, MODULE, NAME, FILE, DOC, MARKERS, STATUS, MESSAGE, DURATION],
        help='define columns in output CSV'
    )
    group.addoption(
        '--csv-add-columns',
        dest='csv_add_columns',
        action='store',
        type=str,
        nargs='+',
        default=[],
        help='add columns to the default set of columns in output CSV'
    )
    group.addoption(
        '--csv-delimiter',
        dest='csv_delimiter',
        action='store',
        metavar='delimiter character',
        default=',',
        help='delimiter character to use in CSV files'
    )
    group.addoption(
        '--csv-quote-char',
        dest='csv_quote_char',
        action='store',
        metavar='quoting character',
        default='"',
        help='quoting character to use in CSV files'
    )


def pytest_addhooks(pluginmanager):
    pluginmanager.add_hookspecs(_hooks)


def pytest_configure(config):
    csv_path = config.option.csv_path
    if csv_path:
        columns_registry = dict(BUILTIN_COLUMNS_REGISTRY)
        config.hook.pytest_csv_register_columns(columns=columns_registry)

        # TODO improve error handling, if user puts a wrong column name it will raise a KeyError
        columns = OrderedDict((column_id, columns_registry[column_id.strip()])
                              for column_ids in config.option.csv_columns + config.option.csv_add_columns
                              for column_id in column_ids.split(','))

        config._csv_reporter = CSVReporter(csv_path=csv_path,
                                           columns=columns,
                                           delimiter=config.option.csv_delimiter,
                                           quote_char=config.option.csv_quote_char)
        config.pluginmanager.register(config._csv_reporter)


def pytest_unconfigure(config):
    csv_reporter = getattr(config, '_csv_reporter', None)
    if csv_reporter:
        del config._csv_reporter
        config.pluginmanager.unregister(csv_reporter)



# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------

ID = 'id'
MODULE = 'module'
CLASS = 'class'
FUNCTION = 'function'
NAME = 'name'
FILE = 'file'
DOC = 'doc'
STATUS = 'status'
SUCCESS = 'success'
DURATION = 'duration'
DURATION_FORMATTED = 'duration_formatted'
MESSAGE = 'message'
MARKERS = 'markers'
MARKERS_WITH_ARGS = 'markers_with_args'
MARKERS_AS_COLUMNS = 'markers_as_columns'
MARKERS_ARGS_AS_COLUMNS = 'markers_args_as_columns'
PARAMETERS = 'parameters'
PARAMETERS_AS_COLUMNS = 'parameters_as_columns'
PROPERTIES = 'properties'
PROPERTIES_AS_COLUMNS = 'properties_as_columns'
HOST = 'host'
USER = 'user'
SYSTEM = 'system'
SYSTEM_RELEASE = 'system_release'
SYSTEM_VERSION = 'system_version'
PYTHON_IMPLEMENTATION = 'python_implementation'
PYTHON_VERSION = 'python_version'
WORKING_DIRECTORY = 'working_directory'



# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------

import datetime
import os

import six

from ._ids import *
from ._status import *
from ._utils import parse_node_id, get_test_doc, get_test_args, get_test_markers, format_mark_info, \
    format_mark_info_args


def column_id(item, report):
    yield ID, item.nodeid


def column_module(item, report):
    mod, cls, func, params = parse_node_id(item.nodeid)
    yield MODULE, mod or ''


def column_class(item, report):
    mod, cls, func, params = parse_node_id(item.nodeid)
    yield CLASS, cls or ''


def column_function(item, report):
    mod, cls, func, params = parse_node_id(item.nodeid)
    yield FUNCTION, func or ''


def column_name(item, report):
    mod, cls, func, params = parse_node_id(item.nodeid)
    yield NAME, '%s[%s]' % (func, params) if params else func


def column_file(item, report):
    yield FILE, item.location[0]


def column_doc(item, report):
    yield DOC, get_test_doc(item).strip()


def column_status(item, report):
    if report.passed:
        yield STATUS, XPASSED if hasattr(report, 'wasxfail') else PASSED
    elif report.failed:
        yield STATUS, XFAILED if hasattr(report, 'wasxfail') else ERROR if report.when != 'call' else FAILED
    elif report.skipped:
        yield STATUS, XFAILED if hasattr(report, 'wasxfail') else SKIPPED


def column_success(item, report):
    yield SUCCESS, str(not bool(report.failed))


def column_message(item, report):
    if report.passed:
        if hasattr(report, 'wasxfail'):
            yield MESSAGE, report.wasxfail
        else:
            yield MESSAGE, ''

    elif report.failed:
        if hasattr(report, 'wasxfail'):
            yield MESSAGE, report.wasxfail
        else:
            if hasattr(report.longrepr, 'reprcrash') and hasattr(report.longrepr.reprcrash, 'message'):
                yield MESSAGE, report.longrepr.reprcrash.message
            elif isinstance(report.longrepr, six.string_types):
                yield MESSAGE, report.longrepr
            else:
                yield MESSAGE, str(report.longrepr)

    elif report.skipped:
        if hasattr(report, 'wasxfail'):
            yield MESSAGE, report.wasxfail
        else:
            _, _, message = report.longrepr
            yield MESSAGE, message


def column_duration(item, report):
    yield DURATION, str(getattr(report, 'duration', ''))


def column_duration_formatted(item, report):
    yield DURATION_FORMATTED, str(datetime.timedelta(seconds=getattr(report, 'duration', 0)))


def column_markers(item, report):
    yield MARKERS, ','.join(format_mark_info(mark, False)
                            for mark in sorted(get_test_markers(item), key=lambda mark: mark.name))


def column_markers_with_args(item, report):
    yield MARKERS, ','.join(format_mark_info(mark, True)
                            for mark in sorted(get_test_markers(item), key=lambda mark: mark.name))


def column_markers_as_columns(item, report):
    for mark in get_test_markers(item):
        yield mark.name, format_mark_info_args(mark) or str(True)


def column_markers_args_as_columns(item, report):
    for mark in get_test_markers(item):
        if not mark.args and not mark.kwargs:
            yield mark.name, str(True)
        else:
            for i, arg in enumerate(mark.args):
                yield '%s.%i' % (mark.name, i), str(arg)
            for name, value in six.iteritems(mark.kwargs):
                yield '%s.%s' % (mark.name, name), str(value)


def column_parameters(item, report):
    yield PARAMETERS, ','.join('%s=%s' % (k, v) for k, v in sorted(six.iteritems(get_test_args(item))))


def column_parameters_as_columns(item, report):
    for name, value in six.iteritems(get_test_args(item)):
        yield name, str(value)


def column_properties(item, report):
    yield PROPERTIES, ','.join('%s=%s' % (k, v) for k, v in sorted(getattr(report, 'user_properties', [])))


def column_properties_as_columns(item, report):
    for name, value in sorted(getattr(report, 'user_properties', [])):
        yield name, str(value)


def column_working_directory(item, report):
    yield WORKING_DIRECTORY, os.getcwd()



# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------


def pytest_csv_register_columns(columns):
    """
    Called on plugin initialization.

    Use it to add your own column types (or override the builtin ones).

    For instance, in conftest.py:
    >>> def pytest_csv_register_columns(columns):
    >>>
    >>>     # constant column
    >>>     columns['my_constant_column'] = 'foobar'
    >>>
    >>>     # simple column
    >>>     columns['my_simple_column'] = lambda item, report: {'my column': report.nodeid}
    >>>
    >>>     # a more complex column type that creates several columns in the CSV
    >>>     def my_multiple_columns(item, report):
    >>>         yield 'my column 1', report.nodeid
    >>>         yield 'my column 2', 42
    >>>     columns['my_multiple_columns'] = my_multiple_columns

    Then run pytest with your new column id:

        $ py.test --csv tests.csv --csv-columns id,status,my_constant_column,my_simple_column,my_multiple_columns

    In certain situations, a (failure) line might need to be emitted for cases
    where the original report was lost. For example, when an xdist test runner
    crashes while running the test. In such cases, the column function will be
    called with item=None and an artificially-generated test failure report;
    if they throw an exception, the their column will be empty.

    :param columns: dictionary of (column id, CSVColumn object)
    """


def pytest_csv_written(csv_path):
    """
    Called whenever a CSV file has been written.

    :param csv_path: CSV file path
    """



# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------

import getpass
import re
from itertools import chain

import os
import six

__NODE_ID__ = re.compile(
    r'(?P<module>.+)\.py(?:::(?P<class>[^:]+)(?:::.+)?)?::(?P<function>[^\[]+)(?:\[(?P<params>.*)\])?')


def parse_node_id(node_id):
    match = re.search(__NODE_ID__, node_id)
    if match:
        return match.group('module').replace('/', '.'), \
               match.group('class') or '', \
               match.group('function'), \
               match.group('params') or ''
    raise Exception('Failed parsing pytest node id: "%s"' % node_id)


def get_user():
    try:
        return getpass.getuser()
    except:
        # workaround for https://bugs.python.org/issue32731
        return os.path.basename(os.path.expanduser("~"))


def get_test_doc(item):
    try:
        return item.obj.__doc__ or ''
    except AttributeError:
        return ''


def get_test_args(item):
    return item.callspec.params if hasattr(item, 'callspec') else {}


def get_test_markers(item):
    if hasattr(item, 'iter_markers'):
        return list(item.iter_markers())

    # pytest <3.8 backward compatibility
    from _pytest.mark import MarkInfo
    return [v for v in six.itervalues(item.keywords) if isinstance(v, MarkInfo)]


def format_mark_info(mark, with_args=True):
    if not with_args or (not mark.args and not mark.kwargs):
        return mark.name
    return '%s(%s)' % (mark.name, format_mark_info_args(mark))


def format_mark_info_args(mark):
    return ','.join(chain(
        (str(arg) for arg in mark.args),
        ('%s=%s' % (k, v) for k, v in sorted(six.iteritems(mark.kwargs)))
    ))



# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------

import platform

from ._builtin import *
from ._ids import *
from ._utils import get_user

BUILTIN_COLUMNS_REGISTRY = {
    ID: column_id,
    MODULE: column_module,
    CLASS: column_class,
    FUNCTION: column_function,
    NAME: column_name,
    FILE: column_file,
    DOC: column_doc,
    STATUS: column_status,
    SUCCESS: column_success,
    DURATION: column_duration,
    DURATION_FORMATTED: column_duration_formatted,
    MESSAGE: column_message,
    MARKERS: column_markers,
    MARKERS_WITH_ARGS: column_markers_with_args,
    MARKERS_AS_COLUMNS: column_markers_as_columns,
    MARKERS_ARGS_AS_COLUMNS: column_markers_args_as_columns,
    PARAMETERS: column_parameters,
    PARAMETERS_AS_COLUMNS: column_parameters_as_columns,
    PROPERTIES: column_properties,
    PROPERTIES_AS_COLUMNS: column_properties_as_columns,
    HOST: platform.node(),
    USER: get_user(),
    SYSTEM: platform.system(),
    SYSTEM_RELEASE: platform.release(),
    SYSTEM_VERSION: platform.version(),
    PYTHON_IMPLEMENTATION: platform.python_implementation(),
    PYTHON_VERSION: platform.python_version(),
    WORKING_DIRECTORY: column_working_directory
}



# ----------------------------------------------------------------------
# pytest-csv - https://github.com/nicoulaj/pytest-csv
# copyright (c) 2018-2021 pytest-csv contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ----------------------------------------------------------------------


__prog__ = 'pytest-csv'
__author__ = 'Julien Nicoulaud'
__email__ = 'julien.nicoulaud@gmail.com'
__url__ = 'https://github.com/nicoulaj/pytest-csv'
__copyright__ = 'copyright 2018 pytest-csv contributors'
__license__ = 'GPLv3'
__description__ = 'CSV output for pytest.'
__keywords__ = 'py.test pytest csv tsv report'
__version__ = '3.0.0'
